package client.model;

import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.math.BigInteger;
import java.net.Socket;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.Iterator;
import java.util.LinkedList;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;

import common.Commands;

import client.listener.ServerListener;

public class Communicator implements Runnable {

	private static final int ANONYMOUS_MASK = 0;
	private static final int HELP_DESK_MASK = 1;
	private static final int ON = 1;
	private static final int OFF = 0;

	private boolean exit;	
	private AnonymousUser user;	
	private DataOutputStream toServer;
	private SSLSocket socket;
	private LinkedList<ServerListener> serverListeners;
	private boolean connected;
	private BufferedReader fromServer;
	private boolean encrypted;

	public Communicator() {
		try {
			this.reset();
			this.exit = false;
			this.connected = false;
			this.encrypted = false;
			serverListeners =  new LinkedList<ServerListener>();
		} catch (IOException e) {
			//Socket is null so exception should never be thrown.
		}
	}	

	public void activateTransferMode() throws IOException, InvalidKeyException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException {
		if (connected && user instanceof HelpDeskUser && user.isLoggedIn()) {
			String command = Commands.TRANSFER_MODE + " " + ON + "\n";
			//secureConnection.send(command);
			toServer.writeBytes(command);
		}
	}

	public void deactivateTransferMode() throws IOException, InvalidKeyException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException {
		if (connected && user instanceof HelpDeskUser && user.isLoggedIn()) {
			String command = Commands.TRANSFER_MODE+ " " + OFF + "\n";
			toServer.writeBytes(command);
		}
	}
	
	public void requestUser() throws IOException, InvalidKeyException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException {
		if (connected && user instanceof HelpDeskUser && user.isLoggedIn()) {
			String command = Commands.REQUEST_USER+ " " + "\n";
			toServer.writeBytes(command);
		}
	}

	public void connect(String hostname, int port) throws IOException {
		System.setProperty("javax.net.ssl.trustStore", "truststore.jks");
		System.setProperty("javax.net.ssl.trustStorePassword", "123456");
		socket = (SSLSocket) SSLSocketFactory.getDefault().createSocket(hostname, port);
		socket.setEnabledCipherSuites(socket.getEnabledCipherSuites());
		socket.setUseClientMode(true);
		socket.startHandshake();
		Thread thread = new Thread(this);
		thread.start();
		InputStream inputStream = socket.getInputStream();
		fromServer = new BufferedReader(new InputStreamReader(inputStream));
		toServer = new DataOutputStream(socket.getOutputStream());		
		connected = true;
	}

	public void login(String username) throws IOException, InvalidKeyException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException {
		if (connected) {
			String command = Commands.LOGIN + " " + ANONYMOUS_MASK + " " + 
					username + "\n";
			toServer.writeBytes(command);
		}
	}

	public void login(String username, String password) throws IOException, InvalidKeyException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException {
		if (connected) {
			String command = Commands.LOGIN + " " + HELP_DESK_MASK + " ";
			command += username + " " + password + "\n";
			toServer.writeBytes(command);
		}
	}

	public void send(String message, String to) throws IOException, InvalidKeyException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException{
		if (connected)  {
			String command  = Commands.MESSAGE + " " + to +
					" " + message + "\n";
			toServer.writeBytes(command);
		}
	}

	public void endChat(String id) throws IOException{
		if (connected) {
			String command = Commands.END_CHAT + " " + id + "\n";	
			toServer.writeBytes(command);
		}
	}

	public void reset() throws IOException {
		if (socket != null) {
			socket.close();
			socket = null;
		}
		connected = false;
	}	

	public void getHelpDeskUsers() throws IOException {
		if (connected) {
			String command = Commands.GET_LIST + "\n";
			toServer.writeBytes(command);
		}
	}

	public void addClientListener(ServerListener listener) {
		if (!serverListeners.contains(listener)) {
			serverListeners.add(listener);
		}		
	}

	public void removeClientListener(ServerListener listener) {
		serverListeners.remove(listener);
	}

	public AnonymousUser getUser() {
		return user;
	}

	public void setUser(AnonymousUser user) {
		this.user = user;
	}

	public boolean isConnected() {
		return connected;
	}

	public void setConnectionStatus(boolean status) {
		this.connected = status;
	}

	public void logout() throws IOException {
		if (connected) {
			String command = Commands.LOGOUT + "\n";
			toServer.writeBytes(command);
			socket.close();
			exit = true;
			System.exit(0);
		}
	}

	public void triggerNewUser(String username) {
		Iterator<ServerListener> iterator = serverListeners.iterator(); 
		ServerListener listener;

		while(iterator.hasNext()) {
			listener = iterator.next();
			listener.newUser(username);		
		}
	}

	
	public void triggerRemoveUser(String username) {
		Iterator<ServerListener> iterator = serverListeners.iterator(); 
		ServerListener listener;

		while(iterator.hasNext()) {
			listener = iterator.next();
			listener.removeUser(username);		
		}
	}

	public void transferUser(String toUser, String anonymousId) throws IOException {
		if (connected) {
			String command = Commands.TRANSFER + " " + toUser + " " + anonymousId +"\n";
			toServer.writeBytes(command);
		}
	}

	public void getTransferList() throws IOException {
		if (connected) {
			String command = Commands.TRANSFER_LIST + "\n";
			toServer.writeBytes(command);
		}
	}

	public void run() {

		String line;
		String command;
		String options;
		int index;

		while (!exit) {
			try {
				line = null;
				if (fromServer != null) {
					line = fromServer.readLine();	
				}
				
				if (line != null) {
					index = line.indexOf(" ");
					if (index > -1) {
						command = line.substring(0, index);
						options = line.substring(index+1);
						triggerServerResponse(command, options);
					}
				}

			} catch (IOException e) {
				//Client server communication error
			} catch (Exception e) {
				// Unknown error caught so as to prevent the thread from stopping
			}
		}
	}

	private void triggerServerResponse(String command, String options) {
		Iterator<ServerListener> iterator = serverListeners.iterator(); 
		ServerListener listener;
		while(iterator.hasNext()) {
			listener = iterator.next();
			listener.serverResponse(command, options);		
		}
	}
}
